terminal = {
    cx = 10, --X origin
    cy = 10, --Y origin
    cpw = 530, --total width
    cph = 310, --total height
    bw = 24, --boarder width, minus X button
    bh = 22, --boarder height
    render = true, --if terminal should render
    focus = true, --if terminal has focus
    inputt = "", --terminal text
    command = "", --command to interpret
    curstimer = 0, --cursor blinking timer
    prompt = "\nC:\\> ",
    cursor = "█", --the blinking cursor
    env = "", --lua enviornment
    dragging = { active = false, diffX = 0, diffY = 0 },
}

local utf8 = require("utf8")

function terminal:update(dt)
if self.render == true then
    self.curstimer = self.curstimer - dt*100
    if self.dragging.active then
        self.focus = true
        self.cx = love.mouse.getX() - self.dragging.diffX
        self.cy = love.mouse.getY() - self.dragging.diffY
    end
    
    --check if the terminal is leaving the upper/left part of the screen
    --don't remove this, the terminal renders incorrectly if it's upperleft corner
    --leaves the screen.
    if self.cx <= 4 then self.cx = 4 end
    if self.cy <= 4 then self.cy = 4 end
    --check if the terminal leaves the lower/right side of the screen.
    --this is optional, and can be disabled
    --if self.cx >= love.window.getWidth()-self.cpw then self.cx = love.window.getWidth()-self.cpw-1 end
    --if self.cy >= love.window.getHeight()-self.cph then self.cy = love.window.getHeight()-self.cph-1 end
else end
end

function terminal:mousepress(x, y, button)
if self.render == true then
    --boarder dragging
    if button == "l"
    and ((x > self.cx and x < self.cx + self.cpw-self.bw)
    and (y > self.cy and y < self.cy + self.bh))
    then
        self.dragging.active = true
        self.dragging.diffX = x - self.cx
        self.dragging.diffY = y - self.cy
    end
    
    --lose focus if clicking anywhere that isn't the window
    if button == "l"
    and ((x > self.cx or x < self.cx + self.cpw)
    or (y < self.cy or y > self.cy + self.cph))
    then
        self.focus = false
    end
    
    --get focus if clicking anywhere on the window
    if button == "l"
    and ((x > self.cx and x < self.cx + self.cpw)
    and (y > self.cy and y < self.cy + self.cph))
    then
        self.focus = true
    end
    
    --X button functionality
    if button == "l"
    and ((x > self.cx+self.cpw-self.bw and x < self.cx+self.cpw)
    and (y > self.cy and y < self.cy + self.bh))
    then
        self.render = false
    end
    
else end
end

function terminal:mouserelease(x, y, button)
if self.render == true then
    if button == "l" then self.dragging.active = false end
else end
end

function terminal:interp()
    if self.command ~= "" then
        local pfunc = " function print(x) return x end "
        local out,err = load("return " .. self.command .. " ", self.env, "bt", self.env)
        if out ~= nil then
            self.env = self.env .. self.command .. tostring(out())
            self.inputt = self.inputt .. "\n" .. tostring(out())
        else
            self.inputt = self.inputt .. "\nnil\n" .. err
        end
    end
end

function terminal:textinput(t)
if self.render == true and self.focus == true then
    self.inputt = self.inputt .. t
    self.command = self.command .. t
else end
end

function terminal:keyboard(key)
    if key == "f1" then
        self.render = not self.render
    end
if self.render == true and self.focus == true then
    if key == "backspace" then
        -- get the byte offset to the last UTF-8 character in the string.
        local byteoffset = utf8.offset(self.inputt, -1)

        if byteoffset then
            -- remove the last UTF-8 character.
            -- string.sub operates on bytes rather than UTF-8 characters, so we couldn't do string.sub(text, 1, -2).
            self.inputt = string.sub(self.inputt, 1, byteoffset - 1)
        end
    elseif key == "return" then
        self:interp()
        self.inputt = self.inputt .. self.prompt
        self.command = ""
    end

else
    if key == "." then terminal.cpw = terminal.cpw + 10 end
    if key == "," then if terminal.cpw > 100 then terminal.cpw = terminal.cpw - 10 else end end
    if key == "=" then terminal.cph = terminal.cph + 10 end
    if key == "-" then if terminal.cph > 100 then terminal.cph = terminal.cph - 10 else end end
end
end

function terminal:maketerm()

end

function terminal:spawn(x,y)
    x = self.cx
    y = self.cy
    local pw,ph = self.cpw, self.cph
    local bwidth = 3 --window boarder thickness
    local name = "Command Prompt"
    local str = "InterValue DOS [Version 6.0.1] \nCopyright (c) 1992 InterValue Corporation. All Rights Reserved.\n"
    local scurs = "█"
    local font = love.graphics.newFont("assets/console.ttf", 14)
    local inputfont = love.graphics.newFont("assets/console.ttf", 15)
    local w, h = x+pw,y+ph
    local canvas = love.graphics.newCanvas(w, h)
    local textcanvas = love.graphics.newCanvas(w-bwidth-1, h-bwidth-1)
    
    love.graphics.setCanvas(canvas)
    canvas:clear()
    love.graphics.setFont(font)
    
    --boarder
    love.graphics.setColor(60, 60, 60, 255)
    love.graphics.rectangle('fill', x-bwidth, y-bwidth, w, h)
    
    --terminal
    love.graphics.setColor(20, 20, 20, 255)
    love.graphics.rectangle('fill', x, y, pw-bwidth, ph-bwidth)
    
    --title bar
    love.graphics.setColor(60, 60, 60, 255)
    love.graphics.rectangle('fill', x, y, w, self.bh)
    if self.focus == true then love.graphics.setColor(200, 200, 200, 255)
    else love.graphics.setColor(120, 120, 120, 255) end
    love.graphics.print(name, x+4, y+4)
    
    --X button
    love.graphics.setColor(100, 20, 20, 255)
    love.graphics.rectangle('fill', x+pw-22, y+2, 16, 15)
    love.graphics.setColor(180, 0, 0, 255)
    love.graphics.rectangle('fill', x+pw-21, y+3, 14, 13)
    love.graphics.setColor(200, 100, 100, 255)
    love.graphics.line(x+pw-21, y+3, x+pw-7, y+16)
    love.graphics.line(x+pw-7, y+3, x+pw-21, y+16)
    
    --text drawing
    love.graphics.setCanvas(textcanvas)
    textcanvas:clear()
    love.graphics.setFont(font)
    love.graphics.setColor(200, 200, 200, 255)
    if self.curstimer <= 0 then
        self.curstimer = 100
    elseif self.curstimer <= 60 and self.focus == true then
        self.cursor = scurs
    else
        self.cursor = " "
    end
    local inst = "\n= taller\n- shorter\n. wider\n, thinner\nr spawn/toggle terminal\n"
    --make a generic scrolling method, not explicitly for text scrolling.
    love.graphics.printf(str .. self.prompt .. self.inputt .. self.cursor, x+bwidth+2, y+self.bh+3, self.cpw-2, "left")
    
    love.graphics.reset()
    if self.render == true then
        love.graphics.draw(canvas)
        love.graphics.draw(textcanvas)
    end
    love.graphics.setCanvas()
end
